#ident	"@(#)enet.c	1.5"
/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/cred.h>
#include <sys/dlpi.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/debug.h>
#include <sys/cmn_err.h>
#include <sys/ddi.h>
#include <sys/enet.h>

struct enet {
	queue_t		*e_rq;		/* read queue */
	ulong_t		e_dlstate;	/* data link provider state */
	ushort_t	e_prot;		/* protocol type */
	ushort_t	e_idtype;	/* type of id (bufcall or timeout) */
	ulong_t		e_id;		/* id for bufcalls and timeouts */
};

#define NO_ID		0
#define BUF_ID		1
#define TIME_ID		2

#define ONESEC		1000000		/* in microseconds */

typedef enum retval { DONE, RETRY, ERR } retval_t;

STATIC int enetopen(queue_t *, dev_t *, int, int,
	cred_t *);
STATIC int enetclose(queue_t *, int, cred_t *);
STATIC int enetrsrv(queue_t *);
STATIC int enetwput(queue_t *, mblk_t *);
STATIC int enetwsrv(queue_t *);
STATIC retval_t enet_info(struct enet *);
STATIC retval_t enet_bind(struct enet *, mblk_t *);
STATIC retval_t enet_unbind(struct enet *, mblk_t *);
STATIC retval_t enet_errorack(struct enet *, ulong_t, ulong_t,
	ulong_t);
STATIC int sameaddr(uchar_t *, uchar_t *);
STATIC int isgroupaddr(uchar_t *);
STATIC void enet_send(struct enet *, mblk_t *);
STATIC void enet_wsched(struct enet *, int);
STATIC void enet_wcont(struct enet *);
STATIC void enet_uderr(struct enet *, uchar_t *, ulong_t,
	ulong_t, ulong_t);
STATIC void enet_free(char *);
STATIC void enet_ioctl(queue_t *, mblk_t *);
STATIC void enet_loop(mblk_t *, struct enethdr *, ushort_t);
void enetinit(void);
void enetintr(int);

STATIC uchar_t enet_addr[MACSZ];
STATIC uchar_t bcast_addr[MACSZ] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};
STATIC int enetsndbusy = 0, enetrcvbusy = 0;
STATIC int enetnext = 0;
STATIC frtn_t enetfrtn = {enet_free, 0};
STATIC struct enet enet[MAXENET];

STATIC struct module_info enet_minfo = {
	0x656e, "enet", MINDATA, MAXDATA, 10*MAXDATA, MAXDATA
};
STATIC struct qinit enet_rinit = {
	NULL, enetrsrv, enetopen, enetclose, NULL, &enet_minfo,
	NULL
};
STATIC struct qinit enet_winit = {
	enetwput, enetwsrv, NULL, NULL, NULL, &enet_minfo, NULL
};
struct streamtab enetinfo = {
	&enet_rinit, &enet_winit, NULL, NULL
};

int enetdevflag = 0;	/* required by config */
int enetspurint = 0;	/* spurious interrupts */
int enetrcvlost = 0;	/* input packets dropped */

void
enetinit()
{
	CMD_REG = RECV_OFF;
}

STATIC int
enetopen(queue_t *q, dev_t *devp, int flag, int sflag,
	cred_t *crp)
{
	int i;

	if (sflag == MODOPEN)
		return(EINVAL);
	if (q->q_ptr != NULL)
		return(0);
	if (sflag == CLONEOPEN) {
		for (i = 0; i < MAXENET; i++)
			if (enet[i].e_rq == NULL)
				break;
		if (i >= MAXENET)
			return(ENXIO);
		*devp = makedevice(getemajor(*devp), i);
	} else {
		if ((i = getminor(*devp)) >= MAXENET)
			return(ENXIO);
	}
	enet[i].e_rq = q;
	enet[i].e_idtype = NO_ID;
	enet[i].e_dlstate = DL_UNBOUND;
	q->q_ptr = (caddr_t)&enet[i];
	WR(q)->q_ptr = (caddr_t)&enet[i];
	return(0);
}

STATIC int
enetclose(queue_t *q, int flag, cred_t *crp)
{
	struct enet *ep;

	ep = (struct enet *)q->q_ptr;
	if (ep->e_idtype == TIME_ID)
		untimeout(ep->e_id);
	else if (ep->e_idtype == BUF_ID)
		unbufcall(ep->e_id);
	ep->e_dlstate = DL_UNBOUND;
	ep->e_rq = NULL;
	q->q_ptr = NULL;
	WR(q)->q_ptr = NULL;
	return(0);
}

STATIC int
enetwput(queue_t *q, mblk_t *mp)
{
	switch (mp->b_datap->db_type) {
	case M_FLUSH:
		if (*mp->b_rptr & FLUSHW) {
			flushq(q, FLUSHALL);
			*mp->b_rptr &= ~FLUSHW;
		}
		if (*mp->b_rptr & FLUSHR) {
			flushq(RD(q), FLUSHALL);
			qreply(q, mp);
		} else {
			freemsg(mp);
		}
		break;

	case M_IOCTL:
		enet_ioctl(q, mp);
		break;

	case M_PROTO:
	case M_PCPROTO:
		putq(q, mp);
		break;

	default:
		freemsg(mp);
		break;
	}
	return(0);
}

STATIC int
enetwsrv(queue_t *q)
{
	mblk_t *mp;
	struct enet *ep;
	int s;
	ulong_t err, prim;
	union DL_primitives *dlp;

	ep = (struct enet *)q->q_ptr;
	if (ep == NULL)
		return(0);
	while ((mp = getq(q)) != NULL) {
		ASSERT((mp->b_datap->db_type == M_PROTO) ||
		  (mp->b_datap->db_type == M_PCPROTO));
		err = 0;
		dlp = (union DL_primitives *)mp->b_rptr;
		prim = dlp->dl_primitive;
		switch (prim) {
		case DL_UNITDATA_REQ:
			s = splstr();
			if (enetsndbusy)
				putbq(q, mp);
			else
				enet_send(ep, mp);
			splx(s);
			return(0);		/* interrupt will reschedule us */

		case DL_INFO_REQ:
			if (enet_info(ep) != DONE)
				cmn_err(CE_WARN,
				  "enet: can't respond to DL_INFO_REQ");
			freemsg(mp);
			break;

		case DL_BIND_REQ:
			if (enet_bind(ep, mp) == RETRY) {
				putbq(q, mp);
				return(0);
			}
			break;

		case DL_UNBIND_REQ:
			if (enet_unbind(ep, mp) == RETRY) {
				putbq(q, mp);
				return(0);
			}
			break;

		case DL_ATTACH_REQ:
		case DL_DETACH_REQ:
		case DL_SUBS_BIND_REQ:
		case DL_ENABMULTI_REQ:
		case DL_DISABMULTI_REQ:
		case DL_PROMISCON_REQ:
		case DL_PROMISCOFF_REQ:
		case DL_UDQOS_REQ:
		case DL_CONNECT_REQ:
		case DL_CONNECT_RES:
		case DL_TOKEN_REQ:
		case DL_DISCONNECT_REQ:
		case DL_RESET_REQ:
		case DL_RESET_RES:
		case DL_DATA_ACK_REQ:
		case DL_REPLY_REQ:
		case DL_REPLY_UPDATE_REQ:
		case DL_XID_REQ:
		case DL_TEST_REQ:
			err = DL_NOTSUPPORTED;	/* fall through */

		default:
			if (err == 0)
				err = DL_BADPRIM;
			if (enet_errorack(ep, prim, err, 0) == RETRY) {
				if (mp->b_datap->db_type < QPCTL) {
					putbq(q, mp);
					return(0);
				}
				cmn_err(CE_WARN,
				  "enet: can't generate DL_ERROR_ACK (prim %d)",
				  prim);
			}
			freemsg(mp);
			break;
		}
	}	/* while */
	return(0);
}

STATIC int
enetrsrv(queue_t *q)
{
	mblk_t *mp;

	while ((mp = getq(q)) != NULL) {
		ASSERT(mp->b_datap->db_type == M_PROTO);
		if (canput(q->q_next)) {
			putnext(q, mp);
		} else {
			putbq(q, mp);
			return(0);
		}
	}
	return(0);
}

STATIC retval_t
enet_info(struct enet *ep)
{
	dl_info_ack_t *ackp;
	mblk_t *mp;
	int i;

	i = DL_INFO_ACK_SIZE+TOTADDRSZ+MACSZ;
	if ((mp = allocb(i, BPRI_HI)) == NULL)
		return(ERR);
	mp->b_datap->db_type = M_PCPROTO;
	ackp = (dl_info_ack_t *)mp->b_wptr;
	ackp->dl_primitive = DL_INFO_ACK;
	ackp->dl_max_sdu = MAXDATA;
	ackp->dl_min_sdu = MINDATA;
	ackp->dl_addr_length = TOTADDRSZ;
	ackp->dl_mac_type = DL_ETHER;
	ackp->dl_reserved = 0;
	ackp->dl_current_state = ep->e_dlstate;
	ackp->dl_sap_length = -SAPSZ;
	ackp->dl_service_mode = DL_CLDLS;
	ackp->dl_qos_length = 0;
	ackp->dl_qos_offset = 0;
	ackp->dl_qos_range_length = 0;
	ackp->dl_qos_range_offset = 0;
	ackp->dl_provider_style = DL_STYLE1;
	ackp->dl_version = DL_VERSION_2;
	ackp->dl_brdcst_addr_length = MACSZ;
	ackp->dl_growth = 0;
	mp->b_wptr += DL_INFO_ACK_SIZE;
	if (ep->e_dlstate == DL_IDLE) {
		ackp->dl_addr_offset = DL_INFO_ACK_SIZE;
		for (i = 0; i < MACSZ; i++)
			*mp->b_wptr++ = enet_addr[i];
		*mp->b_wptr++ = (ep->e_prot>>8) & 0xff;
		*mp->b_wptr++ = ep->e_prot & 0xff;
		ackp->dl_brdcst_addr_offset = DL_INFO_ACK_SIZE +
		  TOTADDRSZ;
	} else {
		ackp->dl_addr_offset = 0;
		ackp->dl_brdcst_addr_offset = DL_INFO_ACK_SIZE;
	}
	for (i = 0; i < MACSZ; i++)
		*mp->b_wptr++ = bcast_addr[i];
	putnext(ep->e_rq, mp);
	return(DONE);
}

STATIC retval_t
enet_bind(struct enet *ep, mblk_t *mp)
{
	int i;
	dl_bind_req_t *reqp;
	dl_bind_ack_t *ackp;
	mblk_t *bp;
	ushort_t prot;

	if (ep->e_dlstate != DL_UNBOUND) {
		if (enet_errorack(ep, DL_BIND_REQ, DL_OUTSTATE,	
		  0) == RETRY)
			return(RETRY);
		freemsg(mp);
		return(ERR);
	}
	reqp = (dl_bind_req_t *)mp->b_rptr;
	if (reqp->dl_service_mode != DL_CLDLS) {
		if (enet_errorack(ep, DL_BIND_REQ, DL_UNSUPPORTED,
		  0) == RETRY)
			return(RETRY);
		freemsg(mp);
		return(ERR);
	}
	if (reqp->dl_xidtest_flg != 0) {
		if (enet_errorack(ep, DL_BIND_REQ, DL_NOAUTO,
		  0) == RETRY)
			return(RETRY);
		freemsg(mp);
		return(ERR);
	}
	prot = reqp->dl_sap;
	for (i = 0; i < MAXENET; i++)
		if ((enet[i].e_prot == prot) &&
		  (enet[i].e_dlstate == DL_IDLE)) {
			if (enet_errorack(ep, DL_BIND_REQ, DL_NOADDR,
			  0) == RETRY)
				return(RETRY);
			freemsg(mp);
			return(ERR);
		}
	if ((bp = allocb(DL_BIND_ACK_SIZE+TOTADDRSZ, BPRI_HI)) ==
	  NULL) {
		enet_wsched(ep, DL_BIND_ACK_SIZE+TOTADDRSZ);
		return(RETRY);
	}
	freemsg(mp);
	bp->b_datap->db_type = M_PCPROTO;
	ackp = (dl_bind_ack_t *)bp->b_wptr;
	ackp->dl_primitive = DL_BIND_ACK;
	ackp->dl_sap = prot;
	ackp->dl_addr_length = TOTADDRSZ;
	ackp->dl_addr_offset = DL_BIND_ACK_SIZE;
	ackp->dl_max_conind = 0;
	ackp->dl_xidtest_flg = 0;
	bp->b_wptr += DL_BIND_ACK_SIZE;
	for (i = 0; i < MACSZ; i++)
		*bp->b_wptr++ = enet_addr[i];
	*bp->b_wptr++ = (prot>>8) & 0xff;
	*bp->b_wptr++ = prot & 0xff;
	ep->e_dlstate = DL_IDLE;
	ep->e_prot = prot;
	putnext(ep->e_rq, bp);
	if (!enetrcvbusy)
		CMD_REG = RECV_ON;
	return(DONE);
}	

STATIC retval_t
enet_unbind(struct enet *ep, mblk_t *mp)
{
	mblk_t *bp;
	dl_ok_ack_t *ackp;

	if (ep->e_dlstate != DL_IDLE) {
		if (enet_errorack(ep, DL_UNBIND_REQ, DL_OUTSTATE,
		  0) == RETRY)
			return(RETRY);
		freemsg(mp);
		return(ERR);
	}
	if ((bp = allocb(DL_OK_ACK_SIZE, BPRI_HI)) == NULL) {
		enet_wsched(ep, DL_OK_ACK_SIZE);
		return(RETRY);
	}
	freemsg(mp);
	if (putctl1(ep->e_rq->q_next, M_FLUSH, FLUSHRW) == 0)
		cmn_err(CE_WARN,
		  "enet: can't flush stream on unbind");
	bp->b_datap->db_type = M_PCPROTO;
	ackp = (dl_ok_ack_t *)bp->b_wptr;
	ackp->dl_primitive = DL_OK_ACK;
	ackp->dl_correct_primitive = DL_UNBIND_REQ;
	bp->b_wptr += DL_OK_ACK_SIZE;
	ep->e_dlstate = DL_UNBOUND;
	ep->e_prot = 0;
	putnext(ep->e_rq, bp);
	return(DONE);
}	

STATIC void
enet_send(struct enet *ep, mblk_t *mp)
{
	int i;
	uchar_t *p, *dest;
	dl_unitdata_req_t *reqp;
	struct enethdr e;
	mblk_t *bp;
	long destlen;

	reqp = (dl_unitdata_req_t *)mp->b_rptr;
	p = mp->b_rptr + reqp->dl_dest_addr_offset;
	dest = p;
	destlen = reqp->dl_dest_addr_length;
	if (ep->e_dlstate != DL_IDLE) {
		enet_uderr(ep, dest, destlen, DL_OUTSTATE, 0);
		freemsg(mp);
		return;
	}
	if (reqp->dl_dest_addr_length != TOTADDRSZ) {
		enet_uderr(ep, dest, destlen, DL_BADADDR, 0);
		freemsg(mp);
		return;
	}
	for (i = 0; i < MACSZ; i++) {
		e.dest[i] = *p++;
		e.src[i] = enet_addr[i];
	}
	e.prot.s = *(ushort_t *)p;
	bp = mp;
	mp = mp->b_cont;
	i = msgdsize(mp);
	if ((i > MAXDATA) || (i < MINDATA)) {
		freemsg(bp);
		enet_uderr(ep, dest, destlen, DL_BADDATA, 0);
		return;
	}
	freeb(bp);	/* done with dest */
	if (sameaddr(e.dest, enet_addr)) {
		enet_loop(mp, &e, ep->e_prot);
		return;
	}
	p = (uchar_t *)SEND_BUF_ADDR;	/* copy to board memory */
	bcopy((caddr_t)&e, (caddr_t)p, EHDRSZ);
	p += EHDRSZ;
	for (bp = mp; bp; bp = bp->b_cont) {
		i = bp->b_wptr - bp->b_rptr;
		bcopy((caddr_t)bp->b_rptr, (caddr_t)p, i);
		p += i;
	}
	enetsndbusy = 1;
	CMD_REG = SEND_PKT;
	if (sameaddr(e.dest, bcast_addr)) {
		enet_loop(mp, &e, ep->e_prot);
		return;
	}
	freemsg(mp);
}	

STATIC void
enet_wsched(struct enet *ep, int size)
{
	ep->e_id = bufcall(size, BPRI_HI, enet_wcont, (long)ep);
	if (ep->e_id == 0) {
		ep->e_id = timeout(enet_wcont, (caddr_t)ep,
		  drv_usectohz(ONESEC));
		ep->e_idtype = TIME_ID;
	} else {
		ep->e_idtype = BUF_ID;
	}
}

STATIC void
enet_wcont(struct enet *ep)
{
	if (ep->e_rq != NULL) {
		ep->e_idtype = NO_ID;
		qenable(WR(ep->e_rq));
	}
}

STATIC retval_t
enet_errorack(struct enet *ep, ulong_t prim, ulong_t err,
	ulong_t uerr)
{
	dl_error_ack_t *errp;
	mblk_t *bp;

	if ((bp = allocb(DL_ERROR_ACK_SIZE, BPRI_HI)) == NULL) {
		enet_wsched(ep, DL_ERROR_ACK_SIZE);
		return(RETRY);
	}
	bp->b_datap->db_type = M_PCPROTO;
	errp = (dl_error_ack_t *)bp->b_wptr;
	errp->dl_primitive = DL_ERROR_ACK;
	errp->dl_error_primitive = prim;
	errp->dl_errno = err;
	errp->dl_unix_errno = uerr;
	bp->b_wptr += DL_ERROR_ACK_SIZE;
	putnext(ep->e_rq, bp);
	return(DONE);
}

STATIC void
enet_uderr(struct enet *ep, uchar_t *dest, ulong_t destlen,
	ulong_t err, ulong_t uerr)
{
	dl_uderror_ind_t *errp;
	mblk_t *bp;
	int i;

	i = DL_UDERROR_IND_SIZE+destlen;
	if ((bp = allocb(i, BPRI_HI)) == NULL)
		return;
	bp->b_datap->db_type = M_PROTO;
	errp = (dl_uderror_ind_t *)bp->b_wptr;
	errp->dl_primitive = DL_UDERROR_IND;
	errp->dl_errno = err;
	errp->dl_unix_errno = uerr;
	errp->dl_dest_addr_length = destlen;
	errp->dl_dest_addr_offset = DL_UDERROR_IND_SIZE;
	bp->b_wptr += DL_UDERROR_IND_SIZE;
	bcopy((caddr_t)dest, (caddr_t)bp->b_wptr, destlen);
	bp->b_wptr += destlen;
	putnext(ep->e_rq, bp);
}

void
enetintr(int vec)
{
	dl_unitdata_ind_t *dp;
	uchar_t *p, *protp;
	int i, stat, size;
	mblk_t *mp, *bp;
	ushort_t prot;

	stat = STAT_REG;
	switch (stat) {
	case SEND_DONE:
		enetsndbusy = 0;
		for (i = enetnext; i < MAXENET; i++)
			if ((enet[i].e_dlstate == DL_IDLE) &&
			  (WR(enet[i].e_rq)->q_first != NULL)) {
				qenable(WR(enet[i].e_rq));
				enetnext = i+1;
				return;
			}
		for (i = 0; i < enetnext; i++)
			if ((enet[i].e_dlstate == DL_IDLE) &&
			  (WR(enet[i].e_rq)->q_first != NULL)) {
				qenable(WR(enet[i].e_rq));
				enetnext = i+1;
				return;
			}
		break;

	case RECV_PKT:
		CMD_REG = RECV_OFF;
		size = DL_UNITDATA_IND_SIZE+TOTADDRSZ+TOTADDRSZ;
		if ((mp = allocb(size, BPRI_MED)) == NULL) {
			CMD_REG = RECV_ON;
			enetrcvlost++;
			return;		/* drop packet on floor */
		}
		size = SIZE_REG;
		if ((bp = esballoc(RECV_BUF_ADDR, size, BPRI_MED,
		  &enetfrtn)) == NULL) {
			freeb(mp);
			CMD_REG = RECV_ON;
			enetrcvlost++;
			return;		/* drop packet on floor */
		}
		mp->b_datap->db_type = M_PROTO;
		dp = (dl_unitdata_ind_t *)mp->b_wptr;
		dp->dl_primitive = DL_UNITDATA_IND;
		dp->dl_dest_addr_length = TOTADDRSZ;
		dp->dl_dest_addr_offset = DL_UNITDATA_IND_SIZE;
		dp->dl_src_addr_length = TOTADDRSZ;
		dp->dl_src_addr_offset = DL_UNITDATA_IND_SIZE +
		  TOTADDRSZ;
		p = bp->b_rptr;
		dp->dl_group_address = isgroupaddr(p);
		mp->b_wptr += DL_UNITDATA_IND_SIZE;
		for (i = 0; i < MACSZ; i++)
			*mp->b_wptr++ = *p++;
		protp = mp->b_wptr;
		mp->b_wptr += SAPSZ;
		for (i = 0; i < MACSZ; i++)
			*mp->b_wptr++ = *p++;
		prot = *p << 8;
		*mp->b_wptr++ = *p;
		*protp++ = *p++;
		prot |= *p;
		*mp->b_wptr++ = *p;
		*protp = *p++;
		bp->b_rptr = p;
		mp->b_cont = bp;
		for (i = 0; i < MAXENET; i++)
			if ((enet[i].e_prot == prot) &&
			  (enet[i].e_dlstate == DL_IDLE)) {
				if (canput(enet[i].e_rq)) {
					putq(enet[i].e_rq, mp);
					enetrcvbusy = 1;
					return;
				} else {
					freemsg(mp);
					enetrcvlost++;
				}
				break;
			}
		if (i >= MAXENET)
			freemsg(mp);
		CMD_REG = RECV_ON;
		break;

	default:
		enetspurint++;
		break;
	}
}

STATIC void
enet_ioctl(queue_t *q, mblk_t *mp)
{
	int i;
	uchar_t *p;
	struct iocblk *iocp;
	struct enet *ep;

	iocp = (struct iocblk *)mp->b_rptr;
	ep = (struct enet *)q->q_ptr;
	if (iocp->ioc_count == TRANSPARENT)
		goto nak;
	switch (iocp->ioc_cmd) {
	case SETADDR:
		if (iocp->ioc_count != MACSZ)
			goto nak;
		if (ep->e_dlstate != DL_UNBOUND) {
			iocp->ioc_error = EBUSY;
			goto nak;
		}
		iocp->ioc_error = drv_priv(iocp->ioc_cr);
		if (iocp->ioc_error != 0)
			goto nak;
		p = (uchar_t *)ENET_ADDR;
		for (i = 0; i < MACSZ; i++) {
			enet_addr[i] = *mp->b_cont->b_rptr++;
			*p++ = enet_addr[i];
		}
		CMD_REG = SET_ADDR;
		mp->b_datap->db_type = M_IOCACK;
		iocp->ioc_count = 0;
		break;

	case GETADDR:
		if (iocp->ioc_count != MACSZ)
			goto nak;
		mp->b_cont->b_wptr = mp->b_cont->b_rptr;
		for (i = 0; i < MACSZ; i++)
			*mp->b_cont->b_wptr++ = enet_addr[i];
		mp->b_datap->db_type = M_IOCACK;
		break;

	default:
nak:
		mp->b_datap->db_type = M_IOCNAK;
	}
	qreply(q, mp);
}

STATIC void
enet_loop(mblk_t *mp, struct enethdr *ehp, ushort_t srcsap)
{
	dl_unitdata_ind_t *dp;
	struct enet *ep;
	mblk_t *bp;
	ushort_t prot;
	int i;

	prot = (ehp->prot.b[0]<<8)|ehp->prot.b[1];
	ep = NULL;
	for (i = 0; i < MAXENET; i++)
		if ((enet[i].e_prot == prot) &&
		  (enet[i].e_dlstate == DL_IDLE)) {
			if (canput(enet[i].e_rq))
				ep = &enet[i];
			break;
		}
	if (ep == NULL) {
		freemsg(mp);
		return;
	}
	i = DL_UNITDATA_IND_SIZE+TOTADDRSZ+TOTADDRSZ;
	if ((bp = allocb(i, BPRI_MED)) == NULL) {
		freemsg(mp);
		return;
	}
	bp->b_datap->db_type = M_PROTO;
	dp = (dl_unitdata_ind_t *)bp->b_wptr;
	dp->dl_primitive = DL_UNITDATA_IND;
	dp->dl_dest_addr_length = TOTADDRSZ;
	dp->dl_dest_addr_offset = DL_UNITDATA_IND_SIZE;
	dp->dl_src_addr_length = TOTADDRSZ;
	dp->dl_src_addr_offset = DL_UNITDATA_IND_SIZE +
	  TOTADDRSZ;
	bp->b_wptr += DL_UNITDATA_IND_SIZE;
	for (i = 0; i < MACSZ; i++)
		*bp->b_wptr++ = ehp->dest[i];
	*(ushort_t *)bp->b_wptr = ehp->prot.s;
	bp->b_wptr += SAPSZ;
	for (i = 0; i < MACSZ; i++)
		*bp->b_wptr++ = ehp->src[i];
	*bp->b_wptr++ = (srcsap>>8) & 0xff;
	*bp->b_wptr++ = srcsap & 0xff;
	bp->b_cont = mp;
	putq(ep->e_rq, bp);
}

STATIC void
enet_free(char *dummy)
{
	enetrcvbusy = 0;
	CMD_REG = RECV_ON;
}

STATIC int
sameaddr(uchar_t *a, uchar_t *b)
{
	int i;

	for (i = 0; i < MACSZ; i++)
		if (*a++ != *b++)
			return(0);
	return(1);
}

STATIC int
isgroupaddr(uchar_t *dest)
{
	return(*dest & 0x01);
}
